#
/*
 *	We made two implementations of the ofdmProcessor,
 *	one with an explicit "state", the other one with
 *	the state of the computation implicit
 *	This happens to be the implicit one
 */
#include	"ofdm_processor-a.h"
#include	"gui.h"
//
	ofdm_processor_a::ofdm_processor_a	(virtualInput *theRig,
	                                 DabParams	*p,
	                                 RadioInterface *mr,
	                                 mscHandler 	*msc,
	                                 ficHandler 	*fic,
	                                 int16_t	fillSize,
	                                 RingBuffer<DSPCOMPLEX> *scopeBuffer,
	                                 RingBuffer<DSPCOMPLEX> *iqBuffer,
	                                 int32_t	iBufferSize,
	                                 bool	   lightRun):
	ofdm_processor (theRig, p, mr, msc, fic,
	                fillSize, scopeBuffer, iqBuffer, iBufferSize, lightRun)
	                {
//
//	Most initialization is done in the parent
	inputSize	= 0;
	inputPointer	= inputSize + 1;
	start (QThread::TimeCriticalPriority);
}

	ofdm_processor_a::~ofdm_processor_a	(void) {
	running	= false;	// this will cause an exception to be raised
	while (!isFinished ())
	   usleep (100);
}

//	In this implementation, we implement the states
//	of the state machine by using positions in the code
//	We make the "getNextSampleReady" therefore extremely wise
//
void ofdm_processor_a::getNextSampleReady (void) {
DSPCOMPLEX sample;
DSPCOMPLEX	Z_Tnull;
//
//	finishing is not a real time action, we just wait until
//	a buffer is to be filled again
	if (inputPointer >= inputSize) {
	   volatile int32_t	x = theRig -> Samples ();
	   while (running && (x < 20480)) {	// amount arbitrarily
	      usleep (100);
	      x = theRig -> Samples ();
	   }
	   if (!running)	
	      throw 20;
	   inputPointer = 0;
	   inputSize	= x < 20480 ? x : 20480;
	   inputSize	= theRig -> getSamples (dataBuffer, inputSize);
	   if (dumping)
	      (void)fwrite ((void *)dataBuffer, sizeof (uint8_t),
	                               2 * inputSize, dumpFile);
	}
//
//	OK, we have samples!!
	sample	= DSPCOMPLEX ((float(dataBuffer [2 * inputPointer] - 128)) / 128.0 * gain,
	                      (float(dataBuffer [2 * inputPointer + 1] - 128)) / 128.0 * gain);
	inputPointer ++;
//
//	first: adjust frequency. We need Hz accuracy
	localPhase	-=
	       FreqOffset + Khz (f2Corrector) + frequencyCorrector;
	if (localPhase < 0)
	   localPhase += INPUT_RATE;
	if (localPhase >= INPUT_RATE)
	   localPhase -= INPUT_RATE;
	sample		*= oscillatorTable [localPhase];

	syncBuffer [syncBufferIndex] = sample;
	Z_Tnull		= syncBuffer [(syncBufferIndex - T_null) &
	                                              syncBufferMask];
	syncFilterSum	+= abs (sample) - abs (Z_Tnull);
	signalLevel	+= abs (sample) -
	                   abs (syncBuffer [(syncBufferIndex - 10 * T_null) & syncBufferMask]);
#define	N	8
	if (++ sampleCnt > 2048000 / N) {
	   showSpectrum		(fillSize);
	   show_fineCorrector	(frequencyCorrector);
	   show_coarseCorrector (f2Corrector);
	   sampleCnt = 0;
	}

	syncBufferIndex = (syncBufferIndex + 1) & syncBufferMask;
	tokenLength ++;
	currentSample	= sample;
}
//
//	The main thread, reading sample and trying to locate
//	the ofdm frames
void	ofdm_processor_a::run	(void) {
int32_t		startIndex;
int32_t		i, j;
//	we keep a local index in the syncbuf in case
//	we want to synchronize
int32_t		indexInSyncBuf;
//	fprintf (stderr, "OK we are starting\n");
//
	try {
//	first, we initialize to get the buffers filled
	   for (i = 0; i < 10 * T_null; i ++) 
	      getNextSampleReady ();
//
//	Then we look for a "dip" in the datastream
SyncOnNull:
	   while (syncFilterSum / T_null  > 0.50 * signalLevel / (10 * T_null)) 
	      getNextSampleReady ();
//	OK, we have a null level
	   nullLevel = syncFilterSum / T_null;
//	Now it is waiting for the end
SyncOnEndNULL:
	   while (syncFilterSum / T_null <=  nullLevel) {
	      nullLevel = syncFilterSum / T_null;
	      getNextSampleReady ();
	   }

SyncOnPhase:
//	the syncBufferIndex should point to that
//	and we read up to params -> K samples in the buffer
	indexInSyncBuf	= (syncBufferIndex - 1) & syncBufferMask;
	   for (i = 1; i < params -> K; i ++) { 
	      getNextSampleReady ();
	   }
//
	   startIndex = phaseSynchronizer ->
	                        findIndex (indexInSyncBuf, params -> K);
	   if (startIndex == -1) { // no sync, try again
	      goto SyncOnNull;
	   }
//	here we are synchronized, we need to copy the data we
//	used for synchronization for block 0
	   ofdmBufferIndex	= 0;
	   for (j = startIndex; j < params -> K; j ++)
	      ofdmBuffer [ofdmBufferIndex ++] =
	            syncBuffer [(indexInSyncBuf + j) & syncBufferMask];

	if (((indexInSyncBuf + params -> K) & syncBufferMask) != syncBufferIndex)
	   fprintf (stderr, "Illegal buffering scheme\n");
//	we can assert that indexInSyncBuf + params -> K == syncBufferIndex,
//	so all delayed tokens are consumed
//fprintf (stderr, "%d <--> %d\n", (indexInSyncBuf + j) & syncBufferMask,
//	                      syncBufferIndex);
//	We now have a reference for computing the tokenLength
	   tokenLength = tokenLength - (params -> K - startIndex);
	   if (0.9 * params -> T_F <= tokenLength &&
	                tokenLength <= 1.1 * params -> T_F)
	      avgTokenLength = 0.8 * avgTokenLength + 0.2 * tokenLength;

	   if (++tokenCount > 10) {
	      tokenCount = 0;
	      show_avgTokenLength (avgTokenLength);
	   }
//	and for the next round
	   tokenLength		= params -> K - startIndex;
//	but we are also ready to process block_0

OFDM_PRS:
	   for (i = ofdmBufferIndex; i < (int)T_u; i ++) {
	      getNextSampleReady ();
	      ofdmBuffer [i] = currentSample;
	   }
//
//	block0 will set the phase reference for further decoding
	   ofdmDecoder	-> processBlock_0 (ofdmBuffer);
//
//	after block 0, we will just read in the other (params -> L) blocks
OFDM_SYMBOLS:
//	The first ones are the FIC blocks
	   avgCorr			= 0;
	   for (ofdmSymbolCount = 2;
	        ofdmSymbolCount <= 4; ofdmSymbolCount ++) {
	      FreqCorr	= 0;
	      for (i = 0; i < (int)T_u; i ++) {
	         getNextSampleReady ();
	         ofdmBuffer [i] = currentSample;
	      }
//	and for the frequency correction, the last T_s - T_u samples
	      for (i = (int)T_u; i < (int)T_s; i ++) {
	         getNextSampleReady ();
	         ofdmBuffer [i] = currentSample;
	         FreqCorr += arg (ofdmBuffer [i] *
	                             conj (ofdmBuffer [i - T_u]));
	      }
//	OK, at the end of a block
	      FreqCorr /= (2 * M_PI * (T_s - T_u));
	      avgCorr += FreqCorr;
	      ofdmDecoder -> processToken (ofdmBuffer, ibits, ofdmSymbolCount);
	      my_ficHandler -> process_ficBlock (ibits,
	                                         ofdmSymbolCount);
	   }
//
//	and similar for the (params -> L - 4) MSC blocks
	   for (ofdmSymbolCount = 5;
	        ofdmSymbolCount <= params -> L; ofdmSymbolCount ++) {
	      FreqCorr	= 0;
	      for (i = 0; i < (int)T_u; i ++) {
	         getNextSampleReady ();
	         ofdmBuffer [i] = currentSample;
	      }
	      for (i = (int32_t)T_u; i < (int32_t)T_s; i ++) {
	         getNextSampleReady ();
	         ofdmBuffer [i] = currentSample;
	         FreqCorr += arg (ofdmBuffer [i] *
	                             conj (ofdmBuffer [i - T_u]));
	      }
//	OK, at the end of a block
	      FreqCorr /= (2 * M_PI * (T_s - T_u));
	      avgCorr += FreqCorr;
	      ofdmDecoder -> processToken (ofdmBuffer, ibits, ofdmSymbolCount);
	      if (!lightRun)
	         my_mscHandler -> process_mscBlock (ibits,
	                                            ofdmSymbolCount);
	   }

	   fineCorrector += (avgCorr / params -> L) * params -> carrierDiff;
	   if (fineCorrector > params -> carrierDiff / 2)
	      fineCorrector -= params -> carrierDiff / 2;
	   if (fineCorrector < -params -> carrierDiff / 2)
	      fineCorrector += params -> carrierDiff / 2;
	   frequencyCorrector = 0.9 * frequencyCorrector +
	                        0.1 * fineCorrector;
	   coarseCorrector	   = ofdmDecoder -> coarseCorrector ();
	   if (f2Correction)
	      f2Corrector	+= coarseCorrector * (params -> carrierDiff / 1000);
//
//	and off we go, up to the next frame
	   goto SyncOnNull;
	}
	catch (int e) {
	   fprintf (stderr, "ofdmProcessor will halt\n");
	}
}

